/*
 ============================================================================
 Name        : EtaPhotReggeModel.c
 Author      : Jannes Nys
 Version     : v1.0 November 2016
 Copyright   : 
 Publication : unpublished + eta-MAID
 Description : Photoproduction of a pseudoscalar : gamma + N --> 0- + N'
               Applied to gamma + p --> eta + p
               To change the model, change 'CGLN_Ai'
               The model is based on Regge exchanges  + the MAID solution
 ============================================================================
*/
 
#include "EtaPhotReggeModel.h"


// *********************************************************************************
void MAID_Eta(double s, double complex MltPole[5][10]){
	/* The MAID 2001 solution computed for the reaction
	 *  gamma p --> eta p
	 *  Other reactions are related to isospin
	 *
	 * return double complex MltPole[5][10]
	 * MltPole[1][L] is E_{L+}
	 * MltPole[2][L] is E_{L-}
	 * MltPole[3][L] is M_{L+}
	 * MltPole[4][L] is M_{L-}
	 *
	 * MAXIMUM L is 9
	 * The multipoles are given in 10^-3/mpi+ in the data files and converted here to mfm
	 */


	int L,i,k;
	// solution valid up to W = 2 GeV
	if(sqrt(s)>2.0){
		for(L=0;L<11;L++){
		for(i=0;i<5 ;i++){
			MltPole[i][L] = 0.0;
		}}
		return ;
	}

	int MAXBIN = 514;
	double wval[800]={0}; 
	
	double ReLP[10][4000]={{0},{0}}, ImLP[10][4000]={{0},{0}};
	double complex LEta[10][4000]={{0},{0}};

	double complex Ep[10][515] = {{0},{0}}, Em[10][515] = {{0},{0}};
	double complex Mp[10][515] = {{0},{0}}, Mm[10][515] = {{0},{0}};
	double complex EpEta[10] = {0}, EmEta[10] = {0};
	double complex MpEta[10] = {0}, MmEta[10] = {0};

	FILE *etamaidL[6];	// Load MAID multipoles
		etamaidL[0] = fopen("etamaidL0.txt","r");
		etamaidL[1] = fopen("etamaidL1.txt","r");
		etamaidL[2] = fopen("etamaidL2.txt","r");
		etamaidL[3] = fopen("etamaidL3.txt","r");
		etamaidL[4] = fopen("etamaidL4.txt","r");
		etamaidL[5] = fopen("etamaidL5.txt","r");
	
	/*    W          EL(p)    
 	 	 (MeV)      Re      Im  */
	for( i =1 ; i<4*MAXBIN+1 ; i++){
		for( L = 0 ; L < 6 ; L++){
			fscanf(etamaidL[L],"%lf %lf %lf\n",&wval[i], &ReLP[L][i],&ImLP[L][i]); // proton multipoles
			LEta[L][i] = ReLP[L][i] + I*ImLP[L][i];
		}
	}

	fclose(etamaidL[0]);fclose(etamaidL[1]);fclose(etamaidL[2]);
	fclose(etamaidL[3]);fclose(etamaidL[4]);fclose(etamaidL[5]);

	for( int i=1; i<MAXBIN+1;i++){
	for(L = 0 ; L < 6 ; L++){
		k = i + 0*MAXBIN;  	Ep[L][i] = LEta[L][k];
		k = i + 1*MAXBIN; 	Em[L][i] = LEta[L][k];
		k = i + 2*MAXBIN;	Mp[L][i] = LEta[L][k];
		k = i + 3*MAXBIN;	Mm[L][i] = LEta[L][k];
	}}

	double factor = 1.41377; // 10^-3 * hbar*c/mpi for the conversion
	double w = sqrt(s);
	int ind = floor((w - 1.487)/0.001+1);		// get the closest W index
	double eps = (w-wval[ind]/1000.0)/0.001; 		// get W separation to the current W value
	
	for(L = 0 ; L < 10 ; L++){
		// Linear extrapolation between W bins
		EpEta[L] = (1-eps)*Ep[L][ind] + eps*Ep[L][ind+1];
		EmEta[L] = (1-eps)*Em[L][ind] + eps*Em[L][ind+1];
		MpEta[L] = (1-eps)*Mp[L][ind] + eps*Mp[L][ind+1];
		MmEta[L] = (1-eps)*Mm[L][ind] + eps*Mm[L][ind+1];
		
		// store multipoles for the output
		// remove this factor to get the CGLN Fi in 10^-3/mpi+ as in the MAID page
		MltPole[1][L] = EpEta[L]*factor;
		MltPole[2][L] = EmEta[L]*factor;
		MltPole[3][L] = MpEta[L]*factor;
		MltPole[4][L] = MmEta[L]*factor;
		
	}
	return;
}

// *********************************************************************************
void MltPole2F(double s, double t, double complex MltPole[5][10] ,double complex CGLNF[5]){
	/*
	 * MltPole[1][L] is E_{L+}
	 * MltPole[2][L] is E_{L-}
	 * MltPole[3][L] is M_{L+}
	 * MltPole[4][L] is M_{L-}
	 *
	 * return CGLN Fi in mFm
	 */
	int L,k;

	// kinematics
	double w = sqrt(s);
	double som = 2*MP*MP + META*META;
	double u = -s - t + som;
	double kcm = (s - MP*MP)/(2.0*w);
	double Ex = (s + MP*MP - META*META)/(2*w);
	double qcm = sqrt(Ex*Ex - MP*MP);
	double z = ( s *(t-u) - MP*MP*(META*META - MP*MP))/(4*s*kcm*qcm);	// z = cos theta
	//double tmin = pow(META,4)/4/s - pow(qcm - kcm,2);
	//double tmax = tmin - 4*kcm*qcm;

	double pol1p[10] = {0}; // First derivative of Legendre polynomials
	double pol2p[10] = {0}; // second derivative of Legendre polynomials
	pol1p[1] = 1.0;
	pol1p[2] = 3*z;
	pol1p[3] = (-3 + 15*z*z)/2.0;
	pol1p[4] = (-60*z + 140*z*z*z)/8;
	pol1p[5] = (15 - 210*z*z + 315*z*z*z*z)/8;
	pol1p[6] = (210*z - 1260*z*z*z + 1386*pow(z,5))/16;
	pol1p[7] = (-35 + 945*z*z - 3465*pow(z,4) + 3003*pow(z,6))/16;
	pol1p[8] = (-2520*z + 27720*z*z*z - 72072*pow(z,5) + 51480*pow(z,7))/128;
	pol1p[9] = (315 - 13860*z*z + 90090*pow(z,4) - 180180*pow(z,6) + 109395*pow(z,8))/128;

	pol2p[2] = 3;
	pol2p[3] = 15*z;
	pol2p[4] = (-60+420*z*z)/8;
	pol2p[5] = (-420*z + 1260*z*z*z)/8.0;
	pol2p[6] = (210 - 3780*z*z + 6930*pow(z,4))/16.0;
	pol2p[7] = (1890*z - 13860*z*z*z + 18018*pow(z,5))/16.0;
	pol2p[8] = (-2520 + 83160*z*z - 360360*pow(z,4) + 360360*pow(z,6) )/128;
	pol2p[9] = ((-27720*z + 360360*z*z*z - 1081080*pow(z,5) + 875160*pow(z,7)))/128;

	CGLNF[1] = 0; CGLNF[2] = 0; CGLNF[3] = 0; CGLNF[4] = 0;
	for( L=0; L<10;L++){
		k = L-1; if(L==0){k = 0;}
		/*
		* MltPole[1][L] is E_{L+}
		* MltPole[2][L] is E_{L-}
	    * MltPole[3][L] is M_{L+}
		* MltPole[4][L] is M_{L-}
		*/
		CGLNF[1] = CGLNF[1] + (L*MltPole[3][L] + MltPole[1][L])*pol1p[L+1];
		CGLNF[1] = CGLNF[1] + ( (L+1)*MltPole[4][L] + MltPole[2][L] )*pol1p[k];
		CGLNF[2] = CGLNF[2] + ( (L+1)*MltPole[3][L] +L*MltPole[4][L])*pol1p[L];
		CGLNF[3] = CGLNF[3] + (MltPole[1][L] - MltPole[3][L])*pol2p[L+1];
		CGLNF[3] = CGLNF[3] + (MltPole[2][L] + MltPole[4][L])*pol2p[L-1];
		CGLNF[4] = CGLNF[4] + (MltPole[3][L] - MltPole[1][L] - MltPole[4][L] - MltPole[2][L])*pol2p[L];
	}
	
	return;
}

// *********************************************************************************
double EtaPhotCS_S(double W,double theta){
/*
 * compute the differential cross section from s-channel helicities
 * in micro barns/GeV^2
 * the amplitudes (helicity amplitudes A(=8 pi W H)) come in GeV.mFm and a barn = 100 Fm^2
 * return dSig/dt in micro bar.GeV^-2
 */
	double mass[5] = {0.0, 0.0, MP, META, MP};	// mass = { - , Ma, Mb, Mc, Md}
												// reaction:  a + b --> c + d
	double pa[4], pb[4], pc[4], pd[4];
	int hel[4] = {2,1,0,1};						// hel = {1,+,0,+} (x2)
        double complex res[4] ;
	double sig;
        kin2to2(W, theta, mass, pa, pb, pc, pd);

	res[0] = EtaPhotAmpS(pa, pb, pc, hel);		// hel = {1,+,0,+} (x2)

	hel[1] = -1;								// hel = {1,-,0,+} (x2)
	res[1] = EtaPhotAmpS(pa, pb, pc, hel);

	hel[3] = -1;								// hel = {1,-,0,-} (x2)
	res[2] = EtaPhotAmpS(pa, pb, pc, hel);

	hel[1] = +1;								// hel = {1,+,0,-} (x2)
	res[3] = EtaPhotAmpS(pa, pb, pc, hel);

//        printf("%lf\n",Cos2T(W,cos(theta),mass));
//        printf(" %lf \t %lf \t %lf \t %lf \n",creal(res[0]),creal(res[1]),creal(res[2]),creal(res[3]));
//        printf(" %lf \t %lf \t %lf \t %lf \n",cimag(res[0]),cimag(res[1]),cimag(res[2]),cimag(res[3]));

	// amps are the T hel
	sig = pow(cabs(res[0]),2) + pow(cabs(res[1]),2) + pow(cabs(res[2]),2) + pow(cabs(res[3]),2);
        sig = sig/32.0/M_PI/(W*W-MP*MP)/(W*W-MP*MP)/100.0 ;
	
	return sig;
}


// *********************************************************************************
double EtaPhotBeamAs_S(double W, double theta){
/*
 * compute the beam asymptotic asymmetry from s-channel helicities
 */


  double mass[5] = {0.0, 0.0, MP, META, MP};
  double pa[4], pb[4], pc[4], pd[4];
  int hel[4] = {2,1,0,1};						// hel = {1,+,0,+} (x2)
  double BeamSigma;
  double complex  res[4] ;
  double sig;
  kin2to2(W, theta, mass, pa, pb, pc, pd);

  res[0] = EtaPhotAmpS(pa, pb, pc, hel);		                // hel = {1,+,0,+} (x2)

  hel[1] = -1;								// hel = {1,-,0,+} (x2)
  res[1] = EtaPhotAmpS(pa, pb, pc, hel);

  hel[3] = -1;								// hel = {1,-,0,-} (x2)
  res[2] = EtaPhotAmpS(pa, pb, pc, hel);

  hel[1] = +1;								// hel = {1,+,0,-} (x2)
  res[3] = EtaPhotAmpS(pa, pb, pc, hel);

  sig = cnorm(res[0]) + cnorm(res[1]) + cnorm(res[2]) + cnorm(res[3]);
  sig = sig/2.0;

  // Sigma beam asymmetry
  BeamSigma = creal(res[0]*conj(res[2]) - res[1]*conj(res[3]));
  BeamSigma = BeamSigma/sig;

  return BeamSigma;
}



// *********************************************************************************
double complex EtaPhotAmpS(double pa[4],double pb[4],double pc[4], int hel[4]){
	/* Regge amplitudes for gamma + p --> eta  + p
	 * See arXiv:1505.02321 for the formulas
	 * Amplitudes in the S-CHANNEL !
	 * Helicities are defined in the c.o.m. of (gamma,p)
	 * Inputs:
	 * 		pa, pb, pc 					: four vectors in the c.o.m. frame.
	 * 		hel={mua, mub, muc, mud} 	: 2 x helicities of a,b,c,d
	 * Outputs:
	 * 		ampl						: the amplitude ; complex number
	 * 									  units = GeV.mFm
	 * 									  Fm = Fermi = femtometer = 10^-15 m
	 */
	double complex amp = 0.0;
	double complex Ai[5] = {0}, FiReg[5] = {0}, FiMAID[5] = {0}, Fi[5] = {0};
	double pab[4], pbc[4], pca[4], pd[4];
	struct Kin var;
	double mass[5]={0};
	double complex MltPole[5][10];

	// Only valid for Q^2 = 0 --> photon helicity should be +1 or -1
	if ( abs(hel[0]) != 2 ) return 0.0;

	for(int i=0;i<4;i++){							// compute the Mandelstam invariant
		pab[i] = pa[i] + pb[i];						// from four-vectors
		pbc[i] = pb[i] - pc[i];						//
		pca[i] = pc[i] - pa[i];						//
		pd[i]  = pa[i] + pb[i] - pc[i];				//
	}
	var.s = snorm( pab ) + I*EPS;
	var.t = snorm( pca ) - I*EPS;
	// photon should be real
	mass[1]   = 0.0;
	mass[2]  = sqrt( snorm( pb ) );
	mass[3]  = sqrt( snorm( pc ) );
	mass[4]  = sqrt( snorm( pd ) );

	kinematics(var.s,var.t, mass, &var);	// fill the structure var with all kin. quantities

	// Compute the CGLN Fi for a given model
	// CHANGE HERE TO CHANGE THE MODEL
        // Regge for W > 2 GeV
	if( creal(var.s) > 4.0 ){
		Regge_CGLNAi(var.s,var.t,Ai);			// call the Regge model
		CGLNA2F(var, Ai, FiReg);				// Regge model: Convert CGLN Ai to CGLN Fi
	}
        // MAID for W < 2 GeV
	if( creal(var.s) < 4.0 ){
		MAID_Eta(creal(var.s),  MltPole);						// MAID multipoles for eta photoproduction
		MltPole2F(creal(var.s), creal(var.t), MltPole,FiMAID);	// MAID CGLN Fi for eta photoproduction
	}

	// Combine the two models
	// Note that the Fi are in mFm
	for(int i=0;i<5;i++){
		Fi[i] = FiMAID[i] + FiReg[i];
	}
	
	
	// Test nucleon helicities hel[1] and hel[3]
	// and compute Walker Hi helicity amplitudes
	if ( hel[1] == 1 && hel[3] == 1 ) {
		// hel = {1,+,0,+}
		amp = sqrt(2) * var.sinsh * (Fi[2] + Fi[1] )
			  + 1/sqrt(2) * var.sins * var.cossh * (Fi[3] + Fi[4]);
	}
	else if  ( hel[1] == -1 && hel[3] == -1 ){
		// hel = {1,-,0,-}
		amp = -1/sqrt(2) * var.sins * var.cossh * (Fi[3] + Fi[4]);
	}
	else if ( hel[1] == -1 && hel[3] == 1 ){
		// hel = {1,-,0,+}
		amp = sqrt(2) * var.cossh * (Fi[2] - Fi[1] )
			  + 1/sqrt(2) * var.sins * var.sinsh * (Fi[3] - Fi[4]);
	}
	else if ( hel[1] == 1 && hel[3] == -1 ){
		// hel = {1,+,0,-}
		amp = 1/sqrt(2) * var.sins * var.sinsh * (Fi[3] - Fi[4]);
	}
	else return 0.0 ;
	
	// if negative photon helicity, nucleon flip amplitude changes sign
	if (hel[0] == -2 && hel[1] != hel[3] ) amp = -amp;

	// there is a factor 8*PI*W between Hi and helicity amplitudes
	return amp * 8 * M_PI * csqrt(var.s);
}

// *********************************************************************************
void Regge_CGLNAi(double complex s, double complex t, double complex CGLNA[5]){
	/*
	 * Model for gamma p --> eta p (J. Nys and V. Mathieu, unpublished)
         * Jannes Nys November 2016
	 */

	// Model parameters
	// ================
	// Regge trajectories
	double alphRho[3] = { 1.0-0.9*MRHO*MRHO , 0.9, 0.0 };
        double alphOmega[3] = { 1.0-0.9*MRHO*MRHO , 0.9, 0.0 };
        double alphB[3] = { -0.7*MPI0*MPI0 , 0.7, 0.0 };
        double alphH[3] = { -0.7*MPI0*MPI0 , 0.7, 0.0 }; // degenerate h-b slope
	
	// coupling constants
        double g1rho = 3.43352;		// s-channel nucleon helicity flip
        double g4rho = -1.39707;			// s-channel nucleon helicity non-flip
        double g1omega = 0.115605;			// s-channel nucleon helicity flip
        double g4omega = -3.34608;		// s-channel nucleon helicity non-flip
        double g2b = 4.94633;
        double r1 = 3.00061;
        double r4 = 1.97379;
        double r2 = 6.20409;
        double g2h = g2b/1.5;
	
	double complex Rrho, Romega, Rb, Rh;
	double complex aRho, aOmega, aB, aH;
	
	// calculate the Regge trajectories for the given kinematics
	aRho = alphRho[0] + t*alphRho[1] + t*t*alphRho[2];
	aOmega = alphOmega[0] + t*alphOmega[1] + t*t*alphOmega[2];
	aB = alphB[0] + t*alphB[1] + t*t*alphB[2];
	aH = alphH[0] + t*alphH[1] + t*t*alphH[2];
	
	double complex u = 2.0*MP*MP + META*META - s - t;
        double complex nu = (s-u)/4.0/MP;

	// Regge propagators and signature factors
	// 			normalization	sin and gamma		signature							Regge behaviour
        Rrho = 		alphRho[1] *	realgamma(-aRho,0) *	( cexp(-I*M_PI*aRho)-1.0 )/2.0 * 		cpow(nu,aRho-1);
        Romega = 	alphOmega[1] *	realgamma(-aOmega,0) *	( cexp(-I*M_PI*aOmega)-1.0 )/2.0 * 	cpow(nu,aOmega-1);
        Rb = 		alphB[1] *		realgamma(-aB,0) *		( cexp(-I*M_PI*aB)-1.0 )/2.0 * 		cpow(nu,aB-1);
        Rh = 		alphH[1] *		realgamma(-aH,0) *		( cexp(-I*M_PI*aH)-1.0 )/2.0 *		cpow(nu,aH-1);
	// where -pi/sin(pi*alpha)/Gamma(alpha+1) = Gamma(-alpha)
	
	// Scalar amplitudes:
	// NOTE: here we work with the A_2, while we work with A_2^\prime in the paper
	CGLNA[1] = 0.0;	CGLNA[2] = 0.0;	CGLNA[3] = 0.0;	CGLNA[4] = 0.0;
	
        if(t != 0.){

            // ISOSCALAR
            // Omega contribution
            CGLNA[1] += g1omega*t*Romega*pow(r1,aOmega-1);
            CGLNA[2] += -g1omega*Romega*pow(r1,aOmega-1);
            CGLNA[4] += g4omega*aOmega*Romega*pow(r4,aOmega-1);
            // h contribution
            CGLNA[2] += g2h*Rh*pow(r2,aH-1);

            //// ISOVECTOR
            // Rho contribution
            CGLNA[1] += g1rho*t*Rrho*pow(r1,aRho-1);
            CGLNA[2] += -g1rho*Rrho*pow(r1,aRho-1);
            CGLNA[4] += g4rho*aRho*Rrho*pow(r4,aRho-1);
            // b contribution
            CGLNA[2] += g2b*Rb*pow(r2,aB-1);
        }

        // HEEEEEEEEEEEEEEEEERE
//        printf("%lf\n",creal(t));
//        printf(" %lf \t %lf \t %lf \t %lf \n",creal(CGLNA[1]),creal(CGLNA[2]),creal(CGLNA[3]),creal(CGLNA[4]));
//        printf(" %lf \t %lf \t %lf \t %lf \n",cimag(CGLNA[1]),cimag(CGLNA[2]),cimag(CGLNA[3]),cimag(CGLNA[4]));

	//
        return;
}

// *********************************************************************************
void CGLNF2A(struct Kin var, double complex Ai[5], double complex Fi[5]){
	/*
	 * Compute CGLN Ai(s,t) from CGLN Fi(s,t)
	 * Ai are in GeV^{-2,-4,-3,-3}
	 * Fi are in mFm
	 */
	double complex w;
	double complex E2p, E2m, E4p, E4m;

	//Fi[2] = 0;
	//Fi[3] = 0;
	//Fi[4] = 0;

	w = csqrt(var.s);
	E2p = var.E2s + var.m2 ;
	E2m = var.E2s - var.m2 ;
	E4p = var.E4s + var.m4 ;
	E4m = var.E4s - var.m4 ;

	double complex q = sqrt(E4p*E4m);
	double complex kq = (var.t-var.m3*var.m3)/(2.0*q);
	double complex tmp = (var.t - var.m3*var.m3 + 2*var.s - 2*var.m2*var.m2)/(2*q*(var.m2 + w));

	// kinematical factors
	Fi[1] = Fi[1]/sqrt(E2p*E4p);
	Fi[2] = Fi[2]/sqrt(E2m*E4m);
	Fi[3] = Fi[3]/sqrt(E2m*E4p);
	Fi[4] = Fi[4]/sqrt(E2p*E4m);

	Ai[1] =  (w + var.m4)*Fi[1] + (var.m4 - w)*(w - var.m2)/(w + var.m2)*Fi[2]
			+ (var.m2 + var.m4)/(var.m2 + w)*kq*(Fi[3] + Fi[4]);

	Ai[2] = (w - var.m2)/(w + var.m2)/q*Fi[3] - 1/q*Fi[4];
	Ai[3] = Fi[1] + (w - var.m2)/(w + var.m2)*Fi[2] + tmp*(Fi[3] + Fi[4]);
	Ai[4] = Fi[1] + (w - var.m2)/(w + var.m2)*Fi[2] + kq/(var.m2 + w)*(Fi[3] + Fi[4]);

	// conversion in GeV^-1
	for(int i = 1 ; i<=4 ; i++){
		Ai[i] = Ai[i] * 8.0*M_PI*w/HBC;
		Ai[i] = Ai[i] / (2*var.s - 2*w*var.m2) ;
		Ai[i] = Ai[i] / 1000.0;
        }
	
	return ;
}

// *********************************************************************************
void CGLNA2F(struct Kin var, double complex Ai[5], double complex Fi[5]){
	/*
	 * Compute CGLN Fi(s,t) from CGLN Ai(s,t)
	 * Ai are in GeV^-{2,4,3,3}
	 * Fi are in mFm
	 */
	double complex w;
	double complex E2p, E2m, E4p, E4m;

	w = csqrt(var.s);
	E2p = var.E2s + var.m2 ;
	E2m = var.E2s - var.m2 ;
	E4p = var.E4s + var.m4 ;
	E4m = var.E4s - var.m4 ;

	Fi[1] = (w - var.m2) * Ai[1] + (var.m3*var.m3 - var.t)/2 * (Ai[3] - Ai[4]);
	Fi[1] = Fi[1]  + ( w - var.m2 ) * ( w - var.m4 ) * Ai[4];
	Fi[1] = Fi[1] * csqrt( E2p * E4p ) / ( 8 * M_PI * w);

	Fi[2] = -(w + var.m2) * Ai[1] + (var.m3*var.m3 - var.t)/2 * (Ai[3] - Ai[4]);
	Fi[2] = Fi[2]  + ( w + var.m2 ) * ( w + var.m4 ) * Ai[4];
	Fi[2] = Fi[2] * csqrt( E2m * E4m ) / ( 8 * M_PI * w);

	Fi[3] = (w + var.m2) * (  (w - var.m2)*Ai[2] + Ai[3] - Ai[4] );
	Fi[3] = Fi[3] * csqrt( E2m * E4p) * var.qs  / ( 8 * M_PI * w);

	Fi[4] = (w - var.m2) * ( -(w + var.m2)*Ai[2] + Ai[3] - Ai[4] );
	Fi[4] = Fi[4] * csqrt( E2p * E4m) * var.qs  / ( 8 * M_PI * w);

	// above Fi[] are in GeV^-1
	// conversion in mFm
	// HBC = 0.19732 fm.GeV ; 1b = 100 fm^2 ; HBC^2 = (0.19732)^2 (10 mb).GeV^2
	// 10^4 HBC^2 = 389.35 mub.GeV^2
	double hbarc = 197.32; // 1000*hbar*c
	Fi[1] = hbarc*Fi[1];
	Fi[2] = hbarc*Fi[2];
	Fi[3] = hbarc*Fi[3];
	Fi[4] = hbarc*Fi[4];

	return;
}

// *********************************************************************************
void kin2to2(double W, double theta, double mass[5], double pa[4],double pb[4],double pc[4], double pd[4]){
/*
 * Kinematics of the 2-to-2 reaction, a + b --> c + d
 * Inputs:
 * 		W center of mass energy
 * 		cos cosine of the scattering angle in the center of mass
 * 		mass = {ma,mb,mc,md} vector with the masses of external particles
 * 	Outputs:
 * 		pa, pb, pc, pd are the momenta of the particles
 */
	double Ea, Eb, Ec, Ed; 			// Energies of the particles
	double pi, pf;					// initial and final breakup momenta
	double ma, mb, mc, md;			// masses of the particles

	ma = mass[1]; mb = mass[2]; mc = mass[3]; md = mass[4];

	// Check that inputs are valid
	if( ( ma<0 ) || ( mb<0 ) || ( mc<0 ) || ( md<0 )  ) {
		printf("\n*** Wrong masses in kin2to2 ! *** \n\n");
		return ;
	}
        if( ( W < ma + mb ) || ( W < mc + md )  ) {
		printf("\n*** Wrong total energy in kin2to2 ! *** \n\n");
		return ;
	}

        Ea = ( W*W + ma*ma - mb*mb )/2/W;
        Eb = ( W*W - ma*ma + mb*mb )/2/W;
        Ec = ( W*W + mc*mc - md*md )/2/W;
        Ed = ( W*W - mc*mc + md*md )/2/W;

	pi = sqrt(Ea*Ea - ma*ma);
	pa[0] = Ea; pa[1] = 0; pa[2] = 0; pa[3] = +pi;
	pb[0] = Eb; pb[1] = 0; pb[2] = 0; pb[3] = -pi;

	pf = sqrt(Ec*Ec - mc*mc);
	pc[0] = Ec; pc[1] = +pf*sin(theta); pc[2] = 0; pc[3] = +pf*cos(theta);
	pd[0] = Ed; pd[1] = -pf*sin(theta); pd[2] = 0; pd[3] = -pf*cos(theta);

	return;
}

// *********************************************************************************
double complex Eg2W(double Eg)
{
    // lab to s-channel cm conversion
    return csqrt( MP*MP + 2.0*MP*sqrt(Eg*Eg) );
}


// *********************************************************************************
void kinematics(double complex s, double complex t, double mass[5], struct Kin *var)
{
	double m12, m22, m32, m42; 	// masses squared
	double complex t0, t1, u ;

	var->s = s;
	var->t = t;

	var->m1 = mass[1];
	var->m2 = mass[2];
	var->m3 = mass[3];
	var->m4 = mass[4];

	m12 = mass[1] * mass[1] ;
	m22 = mass[2] * mass[2] ;
	m32 = mass[3] * mass[3] ;
	m42 = mass[4] * mass[4] ;

	var->ks = csqrt( lambda(s, m12, m22) / 4 / s );
	var->qs = csqrt( lambda(s, m32, m42) / 4 / s );
	var->kt = csqrt( lambda(t, m12, m32) / 4 / t );
	var->pt = csqrt( lambda(t, m22, m42) / 4 / t );

	// nuclei energies in s- and t-channel frames
	var->E2s = ( s + m22 - m12 ) / 2 / csqrt(s);
	var->E4s = ( s + m42 - m32 ) / 2 / csqrt(s);
	var->E2t = ( t + m22 - m42 ) / 2 / csqrt(t);
	var->E4t = ( t + m42 - m22 ) / 2 / csqrt(t);

	// t1 = tmin ; t0 = tmax ; t0 < t < t1
	t1 = cpow(m12 - m32 - m22 + m42,2)/(4*s) - (var->ks - var->qs) * (var->ks - var->qs);
	t0 = t1 - 4 * var->ks *var->qs ;
	u = - s - t + m12 + m22 + m32 + m42 ;	// Mandelstam s variable
	var->phi = s * (t - t1) * (t0 - t) ;	// Kibble function

	var->coss = 1 + (t - t1)/(2 * var->qs * var->ks) ;
	//var->cost = (t*(s-u) + (m12-m32) * (m22-m42) )/(4 * t * var->pt * var->kt) ;
	var->cost = (t*(s-u) + (m12-m32) * (m22-m42) )/(csqrt(lambda(var->t,m12,m32) * lambda(var->t,m22,m42) ) ) ;
	var->sins = csqrt( var->phi / s ) / ( 2 * var->qs * var->ks) ;
	var->sint = csqrt( var->phi / t ) / ( 2 * var->kt * var->pt) ;
	var->cossh = csqrt( (1 + var->coss ) / 2 );
	var->sinsh = csqrt( (1 - var->coss ) / 2 );
	var->costh = csqrt( (1 + var->cost ) / 2 );
	var->sinth = csqrt( (1 - var->cost ) / 2 );

	return ;
}

// *********************************************************************************
double Cos2T(double W, double z, double mass[5]){
	/*
         * Compute t from W and cos = z
	 */
	double t;
        double s = W*W;
	double ma, mb, mc, md;			// masses of the particles
	ma = mass[1]; mb = mass[2]; mc = mass[3]; md = mass[4];

	double qi, qf;
	qi = sqrt(lambda(s, ma*ma, mb*mb) /4/s);
	qf = sqrt(lambda(s, mc*mc, md*md) /4/s);
	double som = ma*ma + mb*mb + mc*mc + md*md;

	t = 2*qi*qf*z + som/2 - s/2 - (ma*ma - mb*mb) * (mc*mc - md*md) /2/s ;
	return t;
}

// *********************************************************************************
double T2Cos(double W, double t, double mass[5]){
	/*
         * Compute z = cos from W
	 */
	double z;
        double s = W*W;
	double ma, mb, mc, md;			// masses of the particles
	ma = mass[1]; mb = mass[2]; mc = mass[3]; md = mass[4];

	double qi, qf;
	qi = sqrt(lambda(s, ma*ma, mb*mb) /4/s);
	qf = sqrt(lambda(s, mc*mc, md*md) /4/s);
	double som = ma*ma + mb*mb + mc*mc + md*md;

	z = (t + (ma*ma - mb*mb)*(mc*mc - md*md)/(2.0*s) - som/2.0 + s/2.0 ) /(2*qi*qf);
	return z;
}


// *********************************************************************************
double Coef(double W, double mass[5]){
	/*
	 * dSig/dt = Coef dSig/dOmega
	 * Coef = Pi/qi/qf in GeV^-2
	 */
        double s = W*W;
	double ma, mb, mc, md;			// masses of the particles
	ma = mass[1]; mb = mass[2]; mc = mass[3]; md = mass[4];

	double qi, qf;
	qi = sqrt(lambda(s, ma*ma, mb*mb) /4/s);
	qf = sqrt(lambda(s, mc*mc, md*md) /4/s);
	return M_PI/qi/qf;
}

// *********************************************************************************
double complex lambda(double complex a, double b, double c){
// triangle function
	return a*a + b*b + c*c - 2*(a*b + b*c + c*a);
}

// *********************************************************************************
double snorm(double p[5]){
// Norm squared of a quadri-vector ; p[0] is the energy ; p[1-3] are x,y,z components
	return p[0]*p[0] - ( p[1]*p[1] + p[2]*p[2] + p[3]*p[3] );
}

//// *********************************************************************************
double complex realgamma(double complex z, int OPT)
// Calculate Gamma function for real z
{
	if(OPT == 0){
		return tgamma(creal(z));
	}else{
		return clog(tgamma(creal(z))); // ln gamma
	}
}
		
// *********************************************************************************
double cnorm(double complex x)
{
    return creal(x)*creal(x) + cimag(x)*cimag(x);
}

//double complex cgamma(double complex z,int OPT)
//{
    //double complex g, infini= 1e308+ 0*I; // z0,z1
    //double x0,q1,q2,x,y,th,th1,th2,g0,gr,gi,gr1,gi1;
    //double na,t,x1,y1,sr,si;
    //int j,k;

    //static double a[] = {
        //8.333333333333333e-02,
       //-2.777777777777778e-03,
        //7.936507936507937e-04,
       //-5.952380952380952e-04,
        //8.417508417508418e-04,
       //-1.917526917526918e-03,
        //6.410256410256410e-03,
       //-2.955065359477124e-02,
        //1.796443723688307e-01,
       //-1.39243221690590};

    //x = creal(z);
    //y = cimag(z);
    //if (x > 171) return infini;
    //if ((y == 0.0) && (x == (int)x) && (x <= 0.0))
        //return infini;
    //else if (x < 0.0) {
        //x1 = x;
        //y1 = y;
        //x = -x;
        //y = -y;
    //}
    //x0 = x;
    //if (x <= 7.0) {
        //na = (int)(7.0-x);
        //x0 = x+na;
    //}
    //q1 = sqrt(x0*x0+y*y);
    //th = atan(y/x0);
    //gr = (x0-0.5)*log(q1)-th*y-x0+0.5*log(2.0*M_PI);
    //gi = th*(x0-0.5)+y*log(q1)-y;
    //for (k=0;k<10;k++){
        //t = pow(q1,-1.0-2.0*k);
        //gr += (a[k]*t*cos((2.0*k+1.0)*th));
        //gi -= (a[k]*t*sin((2.0*k+1.0)*th));
    //}
    //if (x <= 7.0) {
        //gr1 = 0.0;
        //gi1 = 0.0;
        //for (j=0;j<na;j++) {
            //gr1 += (0.5*log((x+j)*(x+j)+y*y));
            //gi1 += atan(y/(x+j));
        //}
        //gr -= gr1;
        //gi -= gi1;
    //}
    //if (x1 <= 0.0) {
        //q1 = sqrt(x*x+y*y);
        //th1 = atan(y/x);
        //sr = -sin(M_PI*x)*cosh(M_PI*y);
        //si = -cos(M_PI*x)*sinh(M_PI*y);
        //q2 = sqrt(sr*sr+si*si);
        //th2 = atan(si/sr);
        //if (sr < 0.0) th2 += M_PI;
        //gr = log(M_PI/(q1*q2))-gr;
        //gi = -th1-th2-gi;
        //x = x1;
        //y = y1;
    //}
    //if (OPT == 0) {
        //g0 = exp(gr);
        //gr = g0*cos(gi);
        //gi = g0*sin(gi);
    //}
    //g = gr + I*gi;
    //return g;
//}

